from .base import ModuleTestBase


class TestUrlScan(ModuleTestBase):
    config_overrides = {"modules": {"urlscan": {"urls": True}}}

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://urlscan.io/api/v1/search/?q=blacklanternsecurity.com",
            json={
                "results": [
                    {
                        "task": {
                            "visibility": "public",
                            "method": "api",
                            "domain": "asdf.blacklanternsecurity.com",
                            "apexDomain": "blacklanternsecurity.com",
                            "time": "2023-05-17T01:45:11.391Z",
                            "uuid": "c558b3b3-b274-4339-99ef-301eb043741f",
                            "url": "https://asdf.blacklanternsecurity.com/cna.html",
                        },
                        "stats": {
                            "uniqIPs": 6,
                            "uniqCountries": 3,
                            "dataLength": 926713,
                            "encodedDataLength": 332213,
                            "requests": 22,
                        },
                        "page": {
                            "country": "US",
                            "server": "GitHub.com",
                            "ip": "2606:50c0:8002::153",
                            "mimeType": "text/html",
                            "title": "Vulnerability Program | Black Lantern Security",
                            "url": "https://asdf.blacklanternsecurity.com/cna.html",
                            "tlsValidDays": 89,
                            "tlsAgeDays": 25,
                            "tlsValidFrom": "2023-04-21T19:16:58.000Z",
                            "domain": "asdf.blacklanternsecurity.com",
                            "apexDomain": "blacklanternsecurity.com",
                            "asnname": "FASTLY, US",
                            "asn": "AS54113",
                            "tlsIssuer": "R3",
                            "status": "200",
                        },
                        "_id": "c558b3b3-b274-4339-99ef-301eb043741f",
                        "_score": None,
                        "sort": [1684287911391, "c558b3b3-b274-4339-99ef-301eb043741f"],
                        "result": "https://urlscan.io/api/v1/result/c558b3b3-b274-4339-99ef-301eb043741f/",
                        "screenshot": "https://urlscan.io/screenshots/c558b3b3-b274-4339-99ef-301eb043741f.png",
                    }
                ]
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "https://asdf.blacklanternsecurity.com/cna.html" for e in events), "Failed to detect URL"



import ipaddress
from contextlib import suppress
from cachetools import LFUCache

from bbot.errors import ValidationError
from bbot.core.helpers.dns.engine import all_rdtypes
from bbot.core.helpers.async_helpers import NamedLock
from bbot.modules.base import InterceptModule, BaseModule
from bbot.core.helpers.dns.helpers import extract_targets


class DNSResolve(InterceptModule):
    """
    TODO:
        - scrap event cache in favor of the parent backtracking method
        - don't duplicate resolution on the same host
        - clean up wildcard checking to only happen once, and re-emit/abort if one is detected
        - same thing with main_host_event. we should never be processing two events - only one.
    """

    watched_events = ["*"]
    _priority = 1
    scope_distance_modifier = None

    class HostModule(BaseModule):
        _name = "host"
        _type = "internal"

        def _outgoing_dedup_hash(self, event):
            return hash((event, self.name, event.always_emit))

    @property
    def module_threads(self):
        return self.dns_config.get("threads", 25)

    async def setup(self):
        self.dns_config = self.scan.config.get("dns", {})
        self.dns_disable = self.dns_config.get("disable", False)
        if self.dns_disable:
            return None, "DNS resolution is disabled in the config"

        self.minimal = self.dns_config.get("minimal", False)
        self.dns_search_distance = max(0, int(self.dns_config.get("search_distance", 1)))
        self._emit_raw_records = None

        # event resolution cache
        self._event_cache = LFUCache(maxsize=10000)
        self._event_cache_locks = NamedLock()

        self.host_module = self.HostModule(self.scan)

        return True

    @property
    def _dns_search_distance(self):
        return max(self.scan.scope_search_distance, self.dns_search_distance)

    @property
    def emit_raw_records(self):
        if self._emit_raw_records is None:
            watching_raw_records = any(
                ["RAW_DNS_RECORD" in m.get_watched_events() for m in self.scan.modules.values()]
            )
            omitted_event_types = self.scan.config.get("omit_event_types", [])
            omit_raw_records = "RAW_DNS_RECORD" in omitted_event_types
            self._emit_raw_records = watching_raw_records or not omit_raw_records
        return self._emit_raw_records

    async def filter_event(self, event):
        if (not event.host) or (event.type in ("IP_RANGE",)):
            return False, "event does not have host attribute"
        return True

    async def handle_event(self, event, **kwargs):
        event_is_ip = self.helpers.is_ip(event.host)
        event_host = str(event.host)
        event_host_hash = hash(event_host)

        async with self._event_cache_locks.lock(event_host_hash):
            # first thing we do is check for wildcards
            if not event_is_ip:
                if event.scope_distance <= self.scan.scope_search_distance:
                    await self.handle_wildcard_event(event)

        event_host = str(event.host)
        event_host_hash = hash(event_host)

        # we do DNS resolution inside a lock to make sure we don't duplicate work
        # once the resolution happens, it will be cached so it doesn't need to happen again
        async with self._event_cache_locks.lock(event_host_hash):
            try:
                # try to get from cache
                # the "main host event" is the original parent IP_ADDRESS or DNS_NAME
                main_host_event, dns_tags, event_whitelisted, event_blacklisted = self._event_cache[event_host_hash]
                # dns_tags, dns_children, event_whitelisted, event_blacklisted = self._event_cache[event_host_hash]
            except KeyError:

                main_host_event, dns_tags, event_whitelisted, event_blacklisted, raw_record_events = (
                    await self.resolve_event(event)
                )

                # if we're not blacklisted and we haven't already done it, emit the main host event and all its raw records
                main_host_resolved = getattr(main_host_event, "_resolved", False)
                if not event_blacklisted and not main_host_resolved:
                    if event_whitelisted:
                        self.debug(
                            f"Making {main_host_event} in-scope because it resolves to an in-scope resource (A/AAAA)"
                        )
                        main_host_event.scope_distance = 0
                        await self.handle_wildcard_event(main_host_event)

                    in_dns_scope = -1 < main_host_event.scope_distance < self._dns_search_distance

                    if event != main_host_event:
                        await self.emit_event(main_host_event)
                    for raw_record_event in raw_record_events:
                        await self.emit_event(raw_record_event)

                    # kill runaway DNS chains
                    dns_resolve_distance = getattr(event, "dns_resolve_distance", 0)
                    if dns_resolve_distance >= self.helpers.dns.runaway_limit:
                        self.debug(
                            f"Skipping DNS children for {event} because their DNS resolve distances would be greater than the configured value for this scan ({self.helpers.dns.runaway_limit})"
                        )
                        main_host_event.dns_children = {}

                    # emit DNS children
                    if not self.minimal:
                        in_dns_scope = -1 < event.scope_distance < self._dns_search_distance
                        for rdtype, records in main_host_event.dns_children.items():
                            module = self._make_dummy_module(rdtype)
                            for record in records:
                                parents = main_host_event.get_parents()
                                for e in parents:
                                    e_is_host = e.type in ("DNS_NAME", "IP_ADDRESS")
                                    e_parent_matches = str(e.parent.host) == str(main_host_event.host)
                                    e_host_matches = str(e.data) == str(record)
                                    e_module_matches = str(e.module) == str(module)
                                    if e_is_host and e_parent_matches and e_host_matches and e_module_matches:
                                        self.trace(
                                            f"TRYING TO EMIT ALREADY-EMITTED {record}:{rdtype} CHILD OF {main_host_event}, parents: {parents}"
                                        )
                                        return
                                try:
                                    child_event = self.scan.make_event(
                                        record, "DNS_NAME", module=module, parent=main_host_event
                                    )
                                    child_event.discovery_context = f"{rdtype} record for {event.host} contains {child_event.type}: {child_event.host}"
                                    # if it's a hostname and it's only one hop away, mark it as affiliate
                                    if child_event.type == "DNS_NAME" and child_event.scope_distance == 1:
                                        child_event.add_tag("affiliate")
                                    if in_dns_scope or self.preset.in_scope(child_event):
                                        self.debug(f"Queueing DNS child for {event}: {child_event}")
                                        await self.emit_event(child_event)
                                except ValidationError as e:
                                    self.warning(
                                        f'Event validation failed for DNS child of {main_host_event}: "{record}" ({rdtype}): {e}'
                                    )

                # mark the host as resolved
                main_host_event._resolved = True

                # store results in cache
                self._event_cache[event_host_hash] = main_host_event, dns_tags, event_whitelisted, event_blacklisted

        # abort if the event resolves to something blacklisted
        if event_blacklisted:
            return False, f"it has a blacklisted DNS record"

        # if the event resolves to an in-scope IP, set its scope distance to 0
        if event_whitelisted:
            self.debug(f"Making {event} in-scope because it resolves to an in-scope resource")
            event.scope_distance = 0
            await self.handle_wildcard_event(event)

        # transfer resolved hosts
        event._resolved_hosts = main_host_event._resolved_hosts

        # If the event is unresolved, change its type to DNS_NAME_UNRESOLVED
        if event.type == "DNS_NAME" and "unresolved" in event.tags:
            event.type = "DNS_NAME_UNRESOLVED"

    async def resolve_event(self, event):
        dns_tags = set()
        event_whitelisted = False
        event_blacklisted = False

        main_host_event = self.get_dns_parent(event)
        event_host = str(event.host)
        event_is_ip = self.helpers.is_ip(event.host)

        rdtypes_to_resolve = ()
        if event_is_ip:
            if not self.minimal:
                rdtypes_to_resolve = ("PTR",)
        else:
            if self.minimal:
                rdtypes_to_resolve = ("A", "AAAA", "CNAME")
            else:
                rdtypes_to_resolve = all_rdtypes

        # if missing from cache, do DNS resolution
        queries = [(event_host, rdtype) for rdtype in rdtypes_to_resolve]
        error_rdtypes = []
        raw_record_events = []
        async for (query, rdtype), (answer, errors) in self.helpers.dns.resolve_raw_batch(queries):
            if self.emit_raw_records and rdtype not in ("A", "AAAA", "CNAME", "PTR"):
                raw_record_event = self.make_event(
                    {"host": str(event_host), "type": rdtype, "answer": answer.to_text()},
                    "RAW_DNS_RECORD",
                    parent=main_host_event,
                    tags=[f"{rdtype.lower()}-record"],
                    context=f"{rdtype} lookup on {{event.parent.host}} produced {{event.type}}",
                )
                raw_record_events.append(raw_record_event)
            if errors:
                error_rdtypes.append(rdtype)
            dns_tags.add(f"{rdtype.lower()}-record")
            for _rdtype, host in extract_targets(answer):
                try:
                    main_host_event.dns_children[_rdtype].add(host)
                except KeyError:
                    main_host_event.dns_children[_rdtype] = {host}

        # if there were dns resolution errors, notify the user with tags
        for rdtype in error_rdtypes:
            if rdtype not in main_host_event.dns_children:
                dns_tags.add(f"{rdtype.lower()}-error")

        # if there weren't any DNS children and it's not an IP address, tag as unresolved
        if not main_host_event.dns_children and not event_is_ip:
            dns_tags.add("unresolved")

        # check DNS children against whitelists and blacklists
        for rdtype, children in main_host_event.dns_children.items():
            if event_blacklisted:
                break
            for host in children:
                # whitelisting / blacklisting based on resolved hosts
                if rdtype in ("A", "AAAA", "CNAME"):
                    # having a CNAME to an in-scope resource doesn't make you in-scope
                    if (not event_whitelisted) and rdtype != "CNAME":
                        with suppress(ValidationError):
                            if self.scan.whitelisted(host):
                                event_whitelisted = True
                                dns_tags.add(f"dns-whitelisted-{rdtype.lower()}")
                    # CNAME to a blacklisted resource, means you're blacklisted
                    with suppress(ValidationError):
                        if self.scan.blacklisted(host):
                            dns_tags.add("blacklisted")
                            dns_tags.add(f"dns-blacklisted-{rdtype.lower()}")
                            event_blacklisted = True
                            event_whitelisted = False
                            break

                # check for private IPs
                try:
                    ip = ipaddress.ip_address(host)
                    if ip.is_private:
                        dns_tags.add("private-ip")
                except ValueError:
                    continue

        # add DNS tags to main host
        for tag in dns_tags:
            main_host_event.add_tag(tag)

        # set resolved_hosts attribute
        for rdtype, children in main_host_event.dns_children.items():
            if rdtype in ("A", "AAAA", "CNAME"):
                for host in children:
                    main_host_event._resolved_hosts.add(host)

        return main_host_event, dns_tags, event_whitelisted, event_blacklisted, raw_record_events

    async def handle_wildcard_event(self, event):
        self.debug(f"Entering handle_wildcard_event({event})")
        try:
            event_host = str(event.host)
            # check if the dns name itself is a wildcard entry
            wildcard_rdtypes = await self.helpers.is_wildcard(event_host)
            for rdtype, (is_wildcard, wildcard_host) in wildcard_rdtypes.items():
                if is_wildcard == False:
                    continue
                elif is_wildcard == True:
                    event.add_tag("wildcard")
                    wildcard_tag = "wildcard"
                elif is_wildcard == None:
                    wildcard_tag = "error"

                event.add_tag(f"{rdtype.lower()}-{wildcard_tag}")

            # wildcard event modification (www.evilcorp.com --> _wildcard.evilcorp.com)
            if wildcard_rdtypes and not "target" in event.tags:
                # these are the rdtypes that have wildcards
                wildcard_rdtypes_set = set(wildcard_rdtypes)
                # consider the event a full wildcard if all its records are wildcards
                event_is_wildcard = False
                if wildcard_rdtypes_set:
                    event_is_wildcard = all(r[0] == True for r in wildcard_rdtypes.values())

                if event_is_wildcard:
                    if event.type in ("DNS_NAME",) and not "_wildcard" in event.data.split("."):
                        wildcard_parent = self.helpers.parent_domain(event_host)
                        for rdtype, (_is_wildcard, _parent_domain) in wildcard_rdtypes.items():
                            if _is_wildcard:
                                wildcard_parent = _parent_domain
                                break
                        wildcard_data = f"_wildcard.{wildcard_parent}"
                        if wildcard_data != event.data:
                            self.debug(f'Wildcard detected, changing event.data "{event.data}" --> "{wildcard_data}"')
                            event.data = wildcard_data

        finally:
            self.debug(f"Finished handle_wildcard_event({event})")

    def get_dns_parent(self, event):
        """
        Get the first parent DNS_NAME / IP_ADDRESS of an event. If one isn't found, create it.
        """
        for parent in event.get_parents(include_self=True):
            if parent.host == event.host and parent.type in ("IP_ADDRESS", "DNS_NAME", "DNS_NAME_UNRESOLVED"):
                return parent
        tags = set()
        if "target" in event.tags:
            tags.add("target")
        return self.scan.make_event(
            event.host,
            "DNS_NAME",
            module=self.host_module,
            parent=event,
            context="{event.parent.type} has host {event.type}: {event.host}",
            tags=tags,
        )

    def _make_dummy_module(self, name):
        try:
            dummy_module = self.scan.dummy_modules[name]
        except KeyError:
            dummy_module = self.scan._make_dummy_module(name=name, _type="DNS")
            dummy_module._priority = 4
            dummy_module.suppress_dupes = False
            self.scan.dummy_modules[name] = dummy_module
        return dummy_module

    def _dns_child_dedup_hash(self, parent_host, host, rdtype):
        # we deduplicate NS records by their parent domain
        # because otherwise every DNS_NAME has one, and it gets super messy
        if rdtype == "NS":
            _, parent_domain = self.helpers.split_domain(parent_host)
            return hash(f"{parent_domain}:{host}")
        return hash(f"{parent_host}:{host}:{rdtype}")

    def _main_outgoing_dedup_hash(self, event):
        return hash(f"{event.host}")



from bbot.modules.base import BaseModule


class subdomain_enum(BaseModule):
    """
    A typical free API-based subdomain enumeration module
    Inherited by many other modules including sublist3r, dnsdumpster, etc.
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query an API for subdomains"}

    base_url = "https://api.example.com"

    # set module error state after this many failed requests in a row
    abort_after_failures = 5

    # whether to reject wildcard DNS_NAMEs
    reject_wildcards = "strict"

    # set qsize to 10. this helps combat rate limiting by ensuring that a query doesn't execute
    # until the queue is ready to receive its results
    # we don't use 1 because it causes delays due to the asyncio.sleep; 10 gives us reasonable buffer room
    _qsize = 10

    # how to deduplicate incoming events
    # options:
    #   "highest_parent": dedupe by highest parent (highest parent of www.api.test.evilcorp.com is evilcorp.com)
    #   "lowest_parent": dedupe by lowest parent (lowest parent of www.api.test.evilcorp.com is api.test.evilcorp.com)
    dedup_strategy = "highest_parent"

    @property
    def source_pretty_name(self):
        return f"{self.__class__.__name__} API"

    def _incoming_dedup_hash(self, event):
        """
        Determines the criteria for what is considered to be a duplicate event if `accept_dupes` is False.
        """
        return hash(self.make_query(event)), f"dedup_strategy={self.dedup_strategy}"

    async def handle_event(self, event):
        query = self.make_query(event)
        results = await self.query(query)
        if results:
            for hostname in set(results):
                if hostname:
                    try:
                        hostname = self.helpers.validators.validate_host(hostname)
                    except ValueError as e:
                        self.verbose(e)
                        continue
                    if hostname and hostname.endswith(f".{query}") and not hostname == event.data:
                        await self.emit_event(
                            hostname,
                            "DNS_NAME",
                            event,
                            abort_if=self.abort_if,
                            context=f'{{module}} searched {self.source_pretty_name} for "{query}" and found {{event.type}}: {{event.data}}',
                        )

    async def request_url(self, query):
        url = f"{self.base_url}/subdomains/{self.helpers.quote(query)}"
        return await self.request_with_fail_count(url)

    def make_query(self, event):
        query = event.data
        parents = list(self.helpers.domain_parents(event.data))
        if self.dedup_strategy == "highest_parent":
            parents = list(reversed(parents))
        elif self.dedup_strategy == "lowest_parent":
            pass
        else:
            raise ValueError('self.dedup_strategy attribute must be set to either "highest_parent" or "lowest_parent"')
        for p in parents:
            if self.scan.in_scope(p):
                query = p
                break
        try:
            return ".".join([s for s in query.split(".") if s != "_wildcard"])
        except Exception:
            self.critical(query)
            raise

    def parse_results(self, r, query=None):
        json = r.json()
        if json:
            for hostname in json:
                yield hostname

    async def query(self, query, parse_fn=None, request_fn=None):
        if parse_fn is None:
            parse_fn = self.parse_results
        if request_fn is None:
            request_fn = self.request_url
        try:
            response = await request_fn(query)
            if response is None:
                self.info(f'Query "{query}" failed (no response)')
                return []
            try:
                results = list(parse_fn(response, query))
            except Exception as e:
                if response:
                    self.info(
                        f'Error parsing results for query "{query}" (status code {response.status_code})', trace=True
                    )
                    self.log.trace(repr(response.text))
                else:
                    self.info(f'Error parsing results for "{query}": {e}', trace=True)
                return
            if results:
                return results
            self.debug(f'No results for "{query}"')
        except Exception as e:
            self.info(f"Error retrieving results for {query}: {e}", trace=True)

    async def _is_wildcard(self, query):
        if self.helpers.is_dns_name(query):
            for domain, wildcard_rdtypes in (await self.helpers.is_wildcard_domain(query)).items():
                if any(t in wildcard_rdtypes for t in ("A", "AAAA", "CNAME")):
                    return True
        return False

    async def filter_event(self, event):
        query = self.make_query(event)
        # check if wildcard
        is_wildcard = await self._is_wildcard(query)
        # check if cloud
        is_cloud = False
        if any(t.startswith("cloud-") for t in event.tags):
            is_cloud = True
        # reject if it's a cloud resource and not in our target
        if is_cloud and event not in self.scan.target:
            return False, "Event is a cloud resource and not a direct target"
        # optionally reject events with wildcards / errors
        if self.reject_wildcards:
            if any(t in event.tags for t in ("a-error", "aaaa-error")):
                return False, "Event has a DNS resolution error"
            if self.reject_wildcards == "strict":
                if is_wildcard:
                    return False, "Event is a wildcard domain"
            elif self.reject_wildcards == "cloud_only":
                if is_wildcard and is_cloud:
                    return False, "Event is both a cloud resource and a wildcard domain"
        return True, ""

    async def abort_if(self, event):
        # this helps weed out unwanted results when scanning IP_RANGES and wildcard domains
        if "in-scope" not in event.tags:
            return True
        return False


class subdomain_enum_apikey(subdomain_enum):
    """
    A typical module for authenticated, API-based subdomain enumeration
    Inherited by several other modules including securitytrails, c99.nl, etc.
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "API key"}

    async def setup(self):
        await super().setup()
        return await self.require_api_key()



from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_engine():
    from bbot.core.engine import EngineClient, EngineServer

    counter = 0
    yield_cancelled = False
    yield_errored = False
    return_started = False
    return_finished = False
    return_cancelled = False
    return_errored = False

    class TestEngineServer(EngineServer):

        CMDS = {
            0: "return_thing",
            1: "yield_stuff",
        }

        async def return_thing(self, n):
            nonlocal return_started
            nonlocal return_finished
            nonlocal return_cancelled
            nonlocal return_errored
            try:
                return_started = True
                await asyncio.sleep(n)
                return_finished = True
                return f"thing{n}"
            except asyncio.CancelledError:
                return_cancelled = True
                raise
            except Exception:
                return_errored = True
                raise

        async def yield_stuff(self, n):
            nonlocal counter
            nonlocal yield_cancelled
            nonlocal yield_errored
            try:
                for i in range(n):
                    yield f"thing{i}"
                    counter += 1
                    await asyncio.sleep(0.1)
            except asyncio.CancelledError:
                yield_cancelled = True
                raise
            except Exception:
                yield_errored = True
                raise

    class TestEngineClient(EngineClient):

        SERVER_CLASS = TestEngineServer

        async def return_thing(self, n):
            return await self.run_and_return("return_thing", n)

        async def yield_stuff(self, n):
            async for _ in self.run_and_yield("yield_stuff", n):
                yield _

    test_engine = TestEngineClient()

    # test return functionality
    return_res = await test_engine.return_thing(1)
    assert return_res == "thing1"

    # test async generator
    assert counter == 0
    assert yield_cancelled == False
    yield_res = [r async for r in test_engine.yield_stuff(13)]
    assert yield_res == [f"thing{i}" for i in range(13)]
    assert len(yield_res) == 13
    assert counter == 13

    # test async generator with cancellation
    counter = 0
    yield_cancelled = False
    yield_errored = False
    agen = test_engine.yield_stuff(1000)
    async for r in agen:
        if counter > 10:
            await agen.aclose()
            break
    await asyncio.sleep(5)
    assert yield_cancelled == True
    assert yield_errored == False
    assert counter < 15

    # test async generator with error
    yield_cancelled = False
    yield_errored = False
    agen = test_engine.yield_stuff(None)
    with pytest.raises(BBOTEngineError):
        async for _ in agen:
            pass
    assert yield_cancelled == False
    assert yield_errored == True

    # test return with cancellation
    return_started = False
    return_finished = False
    return_cancelled = False
    return_errored = False
    task = asyncio.create_task(test_engine.return_thing(2))
    await asyncio.sleep(1)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task
    await asyncio.sleep(0.1)
    assert return_started == True
    assert return_finished == False
    assert return_cancelled == True
    assert return_errored == False

    # test return with late cancellation
    return_started = False
    return_finished = False
    return_cancelled = False
    return_errored = False
    task = asyncio.create_task(test_engine.return_thing(1))
    await asyncio.sleep(2)
    task.cancel()
    result = await task
    assert result == "thing1"
    assert return_started == True
    assert return_finished == True
    assert return_cancelled == False
    assert return_errored == False

    # test return with error
    return_started = False
    return_finished = False
    return_cancelled = False
    return_errored = False
    with pytest.raises(BBOTEngineError):
        result = await test_engine.return_thing(None)
    assert return_started == True
    assert return_finished == False
    assert return_cancelled == False
    assert return_errored == True

    await test_engine.shutdown()



import dns


class MockResolver:

    def __init__(self, mock_data=None):
        self.mock_data = mock_data if mock_data else {}
        self.nameservers = ["127.0.0.1"]

    async def resolve_address(self, ipaddr, *args, **kwargs):
        modified_kwargs = {}
        modified_kwargs.update(kwargs)
        modified_kwargs["rdtype"] = "PTR"
        return await self.resolve(str(dns.reversename.from_address(ipaddr)), *args, **modified_kwargs)

    def create_dns_response(self, query_name, rdtype):
        query_name = query_name.strip(".")
        answers = self.mock_data.get(query_name, {}).get(rdtype, [])
        if not answers:
            raise dns.resolver.NXDOMAIN(f"No answer found for {query_name} {rdtype}")

        message_text = f"""id 1234
opcode QUERY
rcode NOERROR
flags QR AA RD
;QUESTION
{query_name}. IN {rdtype}
;ANSWER"""
        for answer in answers:
            message_text += f"\n{query_name}. 1 IN {rdtype} {answer}"

        message_text += "\n;AUTHORITY\n;ADDITIONAL\n"
        message = dns.message.from_text(message_text)
        return message

    async def resolve(self, query_name, rdtype=None):
        if rdtype is None:
            rdtype = "A"
        elif isinstance(rdtype, str):
            rdtype = rdtype.upper()
        else:
            rdtype = str(rdtype.name).upper()

        domain_name = dns.name.from_text(query_name)
        rdtype_obj = dns.rdatatype.from_text(rdtype)

        if "_NXDOMAIN" in self.mock_data and query_name in self.mock_data["_NXDOMAIN"]:
            # Simulate the NXDOMAIN exception
            raise dns.resolver.NXDOMAIN

        try:
            response = self.create_dns_response(query_name, rdtype)
            answer = dns.resolver.Answer(domain_name, rdtype_obj, dns.rdataclass.IN, response)
            return answer
        except dns.resolver.NXDOMAIN:
            return []



from .base import ModuleTestBase


class TestTXT(ModuleTestBase):
    def check(self, module_test, events):
        txt_file = module_test.scan.home / "output.txt"
        with open(txt_file) as f:
            assert f.read().startswith("[SCAN]")



from .base import ModuleTestBase


class TestDNSCAA(ModuleTestBase):
    targets = ["blacklanternsecurity.notreal"]
    modules_overrides = ["dnscaa", "speculate"]
    config_overrides = {
        "scope": {
            "report_distance": 1,
        }
    }

    async def setup_after_prep(self, module_test):
        await module_test.mock_dns(
            {
                "blacklanternsecurity.notreal": {
                    "A": ["127.0.0.11"],
                    "CAA": [
                        '0 iodef "https://caa.blacklanternsecurity.notreal"',
                        '128 iodef "mailto:caa@blacklanternsecurity.notreal"',
                        '0 issue "comodoca.com"',
                        '1 issue "digicert.com; cansignhttpexchanges=yes"',
                        '0 issuewild "letsencrypt.org"',
                        '128 issuewild "pki.goog; cansignhttpexchanges=yes"',
                    ],
                },
                "caa.blacklanternsecurity.notreal": {"A": ["127.0.0.22"]},
                "comodoca.com": {
                    "A": ["127.0.0.33"],
                    "CAA": [
                        '0 iodef "https://caa.comodoca.com"',
                    ],
                },
                "caa.comodoca.com": {"A": ["127.0.0.33"]},
                "digicert.com": {"A": ["127.0.0.44"]},
                "letsencrypt.org": {"A": ["127.0.0.55"]},
                "pki.goog": {"A": ["127.0.0.66"]},
            }
        )

    def check(self, module_test, events):
        assert any(e.type == "DNS_NAME" and e.data == "comodoca.com" for e in events), "Failed to detect CA DNS name"
        assert any(e.type == "DNS_NAME" and e.data == "digicert.com" for e in events), "Failed to detect CA DNS name"
        assert any(
            e.type == "DNS_NAME" and e.data == "letsencrypt.org" for e in events
        ), "Failed to detect CA DNS name"
        assert any(e.type == "DNS_NAME" and e.data == "pki.goog" for e in events), "Failed to detect CA DNS name"
        assert any(
            e.type == "URL_UNVERIFIED" and e.data == "https://caa.blacklanternsecurity.notreal/" for e in events
        ), "Failed to detect URL"
        assert any(
            e.type == "EMAIL_ADDRESS" and e.data == "caa@blacklanternsecurity.notreal" for e in events
        ), "Failed to detect email address"
        # make sure we're not checking CAA records for out-of-scope hosts
        assert not any(str(e.host) == "caa.comodoca.com" for e in events)


class TestDNSCAAInScopeFalse(TestDNSCAA):
    config_overrides = {"scope": {"report_distance": 3}, "modules": {"dnscaa": {"in_scope_only": False}}}

    def check(self, module_test, events):
        assert any(str(e.host) == "caa.comodoca.com" for e in events)



from bbot.modules.base import BaseModule


class dockerhub(BaseModule):
    watched_events = ["SOCIAL", "ORG_STUB"]
    produced_events = ["SOCIAL", "CODE_REPOSITORY", "URL_UNVERIFIED"]
    flags = ["passive", "safe", "code-enum"]
    meta = {
        "description": "Search for docker repositories of discovered orgs/usernames",
        "created_date": "2024-03-12",
        "author": "@domwhewell-sage",
    }

    site_url = "https://hub.docker.com"
    api_url = f"{site_url}/v2"

    scope_distance_modifier = 2

    async def filter_event(self, event):
        if event.type == "SOCIAL":
            if event.data["platform"] != "docker":
                return False, "platform is not docker"
        return True

    async def handle_event(self, event):
        if event.type == "ORG_STUB":
            await self.handle_org_stub(event)
        elif event.type == "SOCIAL":
            await self.handle_social(event)

    async def handle_org_stub(self, event):
        profile_name = event.data
        # docker usernames are case sensitive, so if there are capitalizations we also try a lowercase variation
        profiles_to_check = set([profile_name, profile_name.lower()])
        for p in profiles_to_check:
            api_url = f"{self.api_url}/users/{p}"
            api_result = await self.helpers.request(api_url, follow_redirects=True)
            status_code = getattr(api_result, "status_code", 0)
            if status_code == 200:
                site_url = f"{self.site_url}/u/{p}"
                # emit social event
                await self.emit_event(
                    {"platform": "docker", "url": site_url, "profile_name": p},
                    "SOCIAL",
                    parent=event,
                    context=f"{{module}} tried {event.type} {event.data} and found docker profile ({{event.type}}) at {p}",
                )

    async def handle_social(self, event):
        username = event.data.get("profile_name", "")
        if not username:
            return
        self.verbose(f"Searching for docker images belonging to {username}")
        repos = await self.get_repos(username)
        for repo in repos:
            await self.emit_event(
                {"url": repo},
                "CODE_REPOSITORY",
                tags="docker",
                parent=event,
                context=f"{{module}} found docker image {{event.type}}: {repo}",
            )

    async def get_repos(self, username):
        repos = []
        url = f"{self.api_url}/repositories/{username}?page_size=25&page=" + "{page}"
        agen = self.helpers.api_page_iter(url, json=False)
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code != 200:
                    break
                try:
                    j = r.json()
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                if not j:
                    break
                for item in j.get("results", []):
                    image_name = item.get("name", "")
                    namespace = item.get("namespace", "")
                    if image_name and namespace:
                        repos.append("https://hub.docker.com/r/" + namespace + "/" + image_name)
        finally:
            agen.aclose()
        return repos



import re
from bbot.modules.base import BaseModule


class social(BaseModule):
    watched_events = ["URL_UNVERIFIED"]
    produced_events = ["SOCIAL"]
    meta = {
        "description": "Look for social media links in webpages",
        "created_date": "2023-03-28",
        "author": "@TheTechromancer",
    }
    flags = ["passive", "safe", "social-enum"]

    # platform name : (regex, case_sensitive)
    social_media_platforms = {
        "linkedin": (r"linkedin.com/(?:in|company)/([a-zA-Z0-9-]+)", False),
        "facebook": (r"facebook.com/([a-zA-Z0-9.]+)", False),
        "twitter": (r"twitter.com/([a-zA-Z0-9_]{1,15})", False),
        "github": (r"github.com/([a-zA-Z0-9_-]+)", False),
        "instagram": (r"instagram.com/([a-zA-Z0-9_.]+)", False),
        "youtube": (r"youtube.com/@([a-zA-Z0-9_]+)", False),
        "bitbucket": (r"bitbucket.org/([a-zA-Z0-9_-]+)", False),
        "gitlab": (r"gitlab.(?:com|org)/([a-zA-Z0-9_-]+)", False),
        "discord": (r"discord.gg/([a-zA-Z0-9_-]+)", True),
        "docker": (r"hub.docker.com/[ru]/([a-zA-Z0-9_-]+)", False),
        "huggingface": (r"huggingface.co/([a-zA-Z0-9_-]+)", False),
    }

    scope_distance_modifier = 1

    async def setup(self):
        self.compiled_regexes = {k: (re.compile(v), c) for k, (v, c) in self.social_media_platforms.items()}
        return True

    async def handle_event(self, event):
        for platform, (regex, case_sensitive) in self.compiled_regexes.items():
            for match in regex.finditer(event.data):
                url = match.group()
                profile_name = match.groups()[0]
                if not case_sensitive:
                    url = url.lower()
                    profile_name = profile_name.lower()
                url = f"https://{url}"
                event_data = {"platform": platform, "url": url, "profile_name": profile_name}
                # only emit if the same event isn't already in the parent chain
                if not any([e.type == "SOCIAL" and e.data == event_data for e in event.get_parents()]):
                    social_event = self.make_event(
                        event_data,
                        "SOCIAL",
                        parent=event,
                    )
                    await self.emit_event(
                        social_event,
                        context=f"{{module}} detected {platform} {{event.type}} at {url}",
                    )



from .base import ModuleTestBase


class TestHunt(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "hunt", "excavate"]
    config_overrides = {
        "interactsh_disable": True,
    }

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": '<html><a href="/hackme.php?cipher=xor">ping</a></html>'}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING" and e.data["description"] == "Found potential INSECURE CRYPTOGRAPHY parameter [cipher]"
            for e in events
        )



from .base import ModuleTestBase


class TestRapidDNS(ModuleTestBase):
    web_body = """<th scope="row ">12</th>
<td>asdf.blacklanternsecurity.com</td>
<td><a href="/sameip/asdf.blacklanternsecurity.com.?t=cname#result" target="_blank" title="asdf.blacklanternsecurity.com. same ip website">asdf.blacklanternsecurity.com.</a>"""

    async def setup_after_prep(self, module_test):
        module_test.module.abort_if = lambda e: False
        module_test.httpx_mock.add_response(
            url=f"https://rapiddns.io/subdomain/blacklanternsecurity.com?full=1#result", text=self.web_body
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



import os
import dns
import time
import asyncio
import logging
import traceback
from cachetools import LRUCache
from contextlib import suppress

from bbot.errors import DNSWildcardBreak
from bbot.core.engine import EngineServer
from bbot.core.helpers.async_helpers import NamedLock
from bbot.core.helpers.dns.helpers import extract_targets
from bbot.core.helpers.misc import (
    is_ip,
    rand_string,
    parent_domain,
    domain_parents,
    clean_dns_record,
)


log = logging.getLogger("bbot.core.helpers.dns.engine.server")

all_rdtypes = ["A", "AAAA", "SRV", "MX", "NS", "SOA", "CNAME", "TXT"]


class DNSEngine(EngineServer):

    CMDS = {
        0: "resolve",
        1: "resolve_raw",
        2: "resolve_batch",
        3: "resolve_raw_batch",
        4: "is_wildcard",
        5: "is_wildcard_domain",
        99: "_mock_dns",
    }

    def __init__(self, socket_path, config={}, debug=False):
        super().__init__(socket_path, debug=debug)

        self.config = config
        self.dns_config = self.config.get("dns", {})
        # config values
        self.timeout = self.dns_config.get("timeout", 5)
        self.retries = self.dns_config.get("retries", 1)
        self.abort_threshold = self.dns_config.get("abort_threshold", 50)

        # resolver
        self.resolver = dns.asyncresolver.Resolver()
        self.resolver.rotate = True
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout

        # skip certain queries
        dns_omit_queries = self.dns_config.get("omit_queries", None)
        if not dns_omit_queries:
            dns_omit_queries = []
        self.dns_omit_queries = dict()
        for d in dns_omit_queries:
            d = d.split(":")
            if len(d) == 2:
                rdtype, query = d
                rdtype = rdtype.upper()
                query = query.lower()
                try:
                    self.dns_omit_queries[rdtype].add(query)
                except KeyError:
                    self.dns_omit_queries[rdtype] = {query}

        # wildcard handling
        self.wildcard_ignore = self.dns_config.get("wildcard_ignore", None)
        if not self.wildcard_ignore:
            self.wildcard_ignore = []
        self.wildcard_ignore = tuple([str(d).strip().lower() for d in self.wildcard_ignore])
        self.wildcard_tests = self.dns_config.get("wildcard_tests", 5)
        self._wildcard_cache = dict()
        # since wildcard detection takes some time, This is to prevent multiple
        # modules from kicking off wildcard detection for the same domain at the same time
        self._wildcard_lock = NamedLock()

        self._dns_connectivity_lock = None
        self._last_dns_success = None
        self._last_connectivity_warning = time.time()
        # keeps track of warnings issued for wildcard detection to prevent duplicate warnings
        self._dns_warnings = set()
        self._errors = dict()
        self._debug = self.dns_config.get("debug", False)
        self._dns_cache = LRUCache(maxsize=10000)

        self.filter_bad_ptrs = self.dns_config.get("filter_ptrs", True)

    async def resolve(self, query, **kwargs):
        """Resolve DNS names and IP addresses to their corresponding results.

        This is a high-level function that can translate a given domain name to its associated IP addresses
        or an IP address to its corresponding domain names. It's structured for ease of use within modules
        and will abstract away most of the complexity of DNS resolution, returning a simple set of results.

        Args:
            query (str): The domain name or IP address to resolve.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            set: A set containing resolved domain names or IP addresses.

        Examples:
            >>> results = await resolve("1.2.3.4")
            {"evilcorp.com"}

            >>> results = await resolve("evilcorp.com")
            {"1.2.3.4", "dead::beef"}
        """
        results = set()
        try:
            answers, errors = await self.resolve_raw(query, **kwargs)
            for answer in answers:
                for _, host in extract_targets(answer):
                    results.add(host)
        except BaseException:
            self.log.trace(f"Caught exception in resolve({query}, {kwargs}):")
            self.log.trace(traceback.format_exc())
            raise

        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results

    async def resolve_raw(self, query, **kwargs):
        """Resolves the given query to its associated DNS records.

        This function is a foundational method for DNS resolution in this class. It understands both IP addresses and
        hostnames and returns their associated records in a raw format provided by the dnspython library.

        Args:
            query (str): The IP address or hostname to resolve.
            type (str or list[str], optional): Specifies the DNS record type(s) to fetch. Can be a single type like 'A'
                or a list like ['A', 'AAAA']. If set to 'any', 'all', or '*', it fetches all supported types. If not
                specified, the function defaults to fetching 'A' and 'AAAA' records.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing two lists:
                - list: A list of tuples where each tuple consists of a record type string (like 'A') and the associated
                  raw dnspython answer.
                - list: A list of tuples where each tuple consists of a record type string and the associated error if
                  there was an issue fetching the record.

        Examples:
            >>> await resolve_raw("8.8.8.8")
            ([('PTR', <dns.resolver.Answer object at 0x7f4a47cdb1d0>)], [])

            >>> await resolve_raw("dns.google")
            (<dns.resolver.Answer object at 0x7f4a47ce46d0>, [])
        """
        # DNS over TCP is more reliable
        # But setting this breaks DNS resolution on Ubuntu because systemd-resolve doesn't support TCP
        # kwargs["tcp"] = True
        try:
            query = str(query).strip()
            kwargs.pop("rdtype", None)
            rdtype = kwargs.pop("type", "A")
            if is_ip(query):
                return await self._resolve_ip(query, **kwargs)
            else:
                return await self._resolve_hostname(query, rdtype=rdtype, **kwargs)
        except BaseException:
            self.log.trace(f"Caught exception in resolve_raw({query}, {kwargs}):")
            self.log.trace(traceback.format_exc())
            raise

    async def _resolve_hostname(self, query, **kwargs):
        """Translate a hostname into its corresponding IP addresses.

        This is the foundational function for converting a domain name into its associated IP addresses. It's designed
        for internal use within the class and handles retries, caching, and a variety of error/timeout scenarios.
        It also respects certain configurations that might ask to skip certain types of queries. Results are returned
        in the default dnspython answer object format.

        Args:
            query (str): The hostname to resolve.
            rdtype (str, optional): The type of DNS record to query (e.g., 'A', 'AAAA'). Defaults to 'A'.
            retries (int, optional): The number of times to retry on failure. Defaults to class-wide `retries`.
            use_cache (bool, optional): Whether to check the cache before trying a fresh resolution. Defaults to True.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved IP addresses.
                - list: A list of errors encountered during the resolution process.

        Examples:
            >>> results, errors = await _resolve_hostname("google.com")
            (<dns.resolver.Answer object at 0x7f4a4b2caf50>, [])
        """
        self.debug(f"Resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        rdtype = kwargs.get("rdtype", "A")

        # skip certain queries if requested
        if rdtype in self.dns_omit_queries:
            if any(h == query or query.endswith(f".{h}") for h in self.dns_omit_queries[rdtype]):
                self.debug(f"Skipping {rdtype}:{query} because it's omitted in the config")
                return results, errors

        parent = parent_domain(query)
        retries = kwargs.pop("retries", self.retries)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        parent_hash = hash(f"{parent}:{rdtype}")
        dns_cache_hash = hash(f"{query}:{rdtype}")
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    error_count = self._errors.get(parent_hash, 0)
                    if error_count >= self.abort_threshold:
                        connectivity = await self._connectivity_check()
                        if connectivity:
                            self.log.verbose(
                                f'Aborting query "{query}" because failed {rdtype} queries for "{parent}" ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                            )
                            if parent_hash not in self._dns_warnings:
                                self.log.verbose(
                                    f'Aborting future {rdtype} queries to "{parent}" because error count ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                                )
                            self._dns_warnings.add(parent_hash)
                            return results, errors
                    results = await self._catch(self.resolver.resolve, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                    if parent_hash in self._errors:
                        self._errors[parent_hash] = 0
                break
            except (
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                TimeoutError,
                asyncio.exceptions.TimeoutError,
            ) as e:
                try:
                    self._errors[parent_hash] += 1
                except KeyError:
                    self._errors[parent_hash] = 1
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    break
                tries_left -= 1
                err_msg = (
                    f'DNS error or timeout for {rdtype} query "{query}" ({self._errors[parent_hash]:,} so far): {e}'
                )
                if tries_left > 0:
                    retry_num = (retries + 1) - tries_left
                    self.debug(err_msg)
                    self.debug(f"Retry (#{retry_num}) resolving {query} with kwargs={kwargs}")
                else:
                    self.log.verbose(err_msg)

        if results:
            self._last_dns_success = time.time()
            self.debug(f"Answers for {query} with kwargs={kwargs}: {list(results)}")

        if errors:
            self.debug(f"Errors for {query} with kwargs={kwargs}: {errors}")

        return results, errors

    async def _resolve_ip(self, query, **kwargs):
        """Translate an IP address into a corresponding DNS name.

        This is the most basic function that will convert an IP address into its associated domain name. It handles
        retries, caching, and multiple types of timeout/error scenarios internally. The function is intended for
        internal use and should not be directly called by modules without understanding its intricacies.

        Args:
            query (str): The IP address to be reverse-resolved.
            retries (int, optional): The number of times to retry on failure. Defaults to 0.
            use_cache (bool, optional): Whether to check the cache for the result before attempting resolution. Defaults to True.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved domain names (in default dnspython answer format).
                - list: A list of errors encountered during resolution.

        Examples:
            >>> results, errors = await _resolve_ip("8.8.8.8")
            (<dns.resolver.Answer object at 0x7f4a47cdb1d0>, [])
        """
        self.debug(f"Reverse-resolving {query} with kwargs={kwargs}")
        retries = kwargs.pop("retries", 0)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        results = []
        errors = []
        dns_cache_hash = hash(f"{query}:PTR")
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    results = await self._catch(self.resolver.resolve_address, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                break
            except (
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                TimeoutError,
                asyncio.exceptions.TimeoutError,
            ) as e:
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    self.debug(f"{e} (query={query}, kwargs={kwargs})")
                    break
                else:
                    tries_left -= 1
                    if tries_left > 0:
                        retry_num = (retries + 2) - tries_left
                        self.debug(f"Retrying (#{retry_num}) {query} with kwargs={kwargs}")

        if results:
            self._last_dns_success = time.time()

        return results, errors

    async def resolve_batch(self, queries, threads=10, **kwargs):
        """
        A helper to execute a bunch of DNS requests.

        Args:
            queries (list): List of queries to resolve.
            **kwargs: Additional keyword arguments to pass to `resolve()`.

        Yields:
            tuple: A tuple containing the original query and its resolved value.

        Examples:
            >>> import asyncio
            >>> async def example_usage():
            ...     async for result in resolve_batch(['www.evilcorp.com', 'evilcorp.com']):
            ...         print(result)
            ('www.evilcorp.com', {'1.1.1.1'})
            ('evilcorp.com', {'2.2.2.2'})
        """
        async for (args, _, _), responses in self.task_pool(
            self.resolve, args_kwargs=queries, threads=threads, global_kwargs=kwargs
        ):
            yield args[0], responses

    async def resolve_raw_batch(self, queries, threads=10, **kwargs):
        queries_kwargs = [[q[0], {"type": q[1]}] for q in queries]
        async for (args, kwargs, _), (answers, errors) in self.task_pool(
            self.resolve_raw, args_kwargs=queries_kwargs, threads=threads, global_kwargs=kwargs
        ):
            query = args[0]
            rdtype = kwargs["type"]
            for answer in answers:
                yield ((query, rdtype), (answer, errors))

    async def _catch(self, callback, *args, **kwargs):
        """
        Asynchronously catches exceptions thrown during DNS resolution and logs them.

        This method wraps around a given asynchronous callback function to handle different
        types of DNS exceptions and general exceptions. It logs the exceptions for debugging
        and, in some cases, re-raises them.

        Args:
            callback (callable): The asynchronous function to be executed.
            *args: Positional arguments to pass to the callback.
            **kwargs: Keyword arguments to pass to the callback.

        Returns:
            Any: The return value of the callback function, or an empty list if an exception is caught.

        Raises:
            dns.resolver.NoNameservers: When no nameservers could be reached.
        """
        try:
            return await callback(*args, **kwargs)
        except dns.resolver.NoNameservers:
            raise
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout, TimeoutError):
            self.log.debug(f"DNS query with args={args}, kwargs={kwargs} timed out after {self.timeout} seconds")
            raise
        except dns.exception.DNSException as e:
            self.debug(f"{e} (args={args}, kwargs={kwargs})")
        except Exception as e:
            self.log.warning(f"Error in {callback.__qualname__}() with args={args}, kwargs={kwargs}: {e}")
            self.log.trace(traceback.format_exc())
        return []

    async def is_wildcard(self, query, ips=None, rdtype=None):
        """
        Use this method to check whether a *host* is a wildcard entry

        This can reliably tell the difference between a valid DNS record and a wildcard within a wildcard domain.

        If you want to know whether a domain is using wildcard DNS, use `is_wildcard_domain()` instead.

        Args:
            query (str): The hostname to check for a wildcard entry.
            ips (list, optional): List of IPs to compare against, typically obtained from a previous DNS resolution of the query.
            rdtype (str, optional): The DNS record type (e.g., "A", "AAAA") to consider during the check.

        Returns:
            dict: A dictionary indicating if the query is a wildcard for each checked DNS record type.
                Keys are DNS record types like "A", "AAAA", etc.
                Values are tuples where the first element is a boolean indicating if the query is a wildcard,
                and the second element is the wildcard parent if it's a wildcard.

        Raises:
            ValueError: If only one of `ips` or `rdtype` is specified or if no valid IPs are specified.

        Examples:
            >>> is_wildcard("www.github.io")
            {"A": (True, "github.io"), "AAAA": (True, "github.io")}

            >>> is_wildcard("www.evilcorp.com", ips=["93.184.216.34"], rdtype="A")
            {"A": (False, "evilcorp.com")}

        Note:
            `is_wildcard` can be True, False, or None (indicating that wildcard detection was inconclusive)
        """
        result = {}

        parent = parent_domain(query)
        parents = list(domain_parents(query))

        if rdtype is not None:
            if isinstance(rdtype, str):
                rdtype = [rdtype]
            rdtypes_to_check = rdtype
        else:
            rdtypes_to_check = all_rdtypes

        query_baseline = dict()
        # if the caller hasn't already done the work of resolving the IPs
        if ips is None:
            # then resolve the query for all rdtypes
            queries = [(query, t) for t in rdtypes_to_check]
            async for (query, _rdtype), (answers, errors) in self.resolve_raw_batch(queries):
                answers = extract_targets(answers)
                if answers:
                    query_baseline[_rdtype] = set([a[1] for a in answers])
                else:
                    if errors:
                        self.debug(f"Failed to resolve {query} ({_rdtype}) during wildcard detection")
                        result[_rdtype] = (None, parent)
                        continue
        else:
            # otherwise, we can skip all that
            cleaned_ips = set([clean_dns_record(ip) for ip in ips])
            if not cleaned_ips:
                raise ValueError("Valid IPs must be specified")
            query_baseline[rdtype] = cleaned_ips
        if not query_baseline:
            return result

        # once we've resolved the base query and have IP addresses to work with
        # we can compare the IPs to the ones we have on file for wildcards

        # for every parent domain, starting with the shortest
        try:
            for host in parents[::-1]:
                # make sure we've checked that domain for wildcards
                await self.is_wildcard_domain(host)

                # for every rdtype
                for _rdtype in list(query_baseline):
                    # get the IPs from above
                    query_ips = query_baseline.get(_rdtype, set())
                    host_hash = hash(host)

                    if host_hash in self._wildcard_cache:
                        # then get its IPs from our wildcard cache
                        wildcard_rdtypes = self._wildcard_cache[host_hash]

                        # then check to see if our IPs match the wildcard ones
                        if _rdtype in wildcard_rdtypes:
                            wildcard_ips = wildcard_rdtypes[_rdtype]
                            # if our IPs match the wildcard ones, then ladies and gentlemen we have a wildcard
                            is_wildcard = any(r in wildcard_ips for r in query_ips)

                            if is_wildcard and not result.get(_rdtype, (None, None))[0] is True:
                                result[_rdtype] = (True, host)

                    # if we've reached a point where the dns name is a complete wildcard, class can be dismissed early
                    base_query_rdtypes = set(query_baseline)
                    wildcard_rdtypes_set = set([k for k, v in result.items() if v[0] is True])
                    if base_query_rdtypes and wildcard_rdtypes_set and base_query_rdtypes == wildcard_rdtypes_set:
                        self.log.debug(
                            f"Breaking from wildcard detection for {query} at {host} because base query rdtypes ({base_query_rdtypes}) == wildcard rdtypes ({wildcard_rdtypes_set})"
                        )
                        raise DNSWildcardBreak()

        except DNSWildcardBreak:
            pass

        for _rdtype, answers in query_baseline.items():
            if answers and _rdtype not in result:
                result[_rdtype] = (False, query)

        return result

    async def is_wildcard_domain(self, domain, log_info=False):
        """
        Check whether a given host or its children make use of wildcard DNS entries. Wildcard DNS can have
        various implications, particularly in subdomain enumeration and subdomain takeovers.

        Args:
            domain (str): The domain to check for wildcard DNS entries.
            log_info (bool, optional): Whether to log the result of the check. Defaults to False.

        Returns:
            dict: A dictionary where the keys are the parent domains that have wildcard DNS entries,
            and the values are another dictionary of DNS record types ("A", "AAAA", etc.) mapped to
            sets of their resolved IP addresses.

        Examples:
            >>> is_wildcard_domain("github.io")
            {"github.io": {"A": {"1.2.3.4"}, "AAAA": {"dead::beef"}}}

            >>> is_wildcard_domain("example.com")
            {}
        """
        wildcard_domain_results = {}

        rdtypes_to_check = set(all_rdtypes)

        # make a list of its parents
        parents = list(domain_parents(domain, include_self=True))
        # and check each of them, beginning with the highest parent (i.e. the root domain)
        for i, host in enumerate(parents[::-1]):
            # have we checked this host before?
            host_hash = hash(host)
            async with self._wildcard_lock.lock(host_hash):
                # if we've seen this host before
                if host_hash in self._wildcard_cache:
                    wildcard_domain_results[host] = self._wildcard_cache[host_hash]
                    continue

                self.log.verbose(f"Checking if {host} is a wildcard")

                # determine if this is a wildcard domain

                # resolve a bunch of random subdomains of the same parent
                is_wildcard = False
                wildcard_results = dict()

                rand_queries = []
                for rdtype in rdtypes_to_check:
                    for _ in range(self.wildcard_tests):
                        rand_query = f"{rand_string(digits=False, length=10)}.{host}"
                        rand_queries.append((rand_query, rdtype))

                async for (query, rdtype), (answers, errors) in self.resolve_raw_batch(rand_queries, use_cache=False):
                    answers = extract_targets(answers)
                    if answers:
                        is_wildcard = True
                        if not rdtype in wildcard_results:
                            wildcard_results[rdtype] = set()
                        wildcard_results[rdtype].update(set(a[1] for a in answers))
                        # we know this rdtype is a wildcard
                        # so we don't need to check it anymore
                        with suppress(KeyError):
                            rdtypes_to_check.remove(rdtype)

                self._wildcard_cache.update({host_hash: wildcard_results})
                wildcard_domain_results.update({host: wildcard_results})
                if is_wildcard:
                    wildcard_rdtypes_str = ",".join(sorted([t.upper() for t, r in wildcard_results.items() if r]))
                    log_fn = self.log.verbose
                    if log_info:
                        log_fn = self.log.info
                    log_fn(f"Encountered domain with wildcard DNS ({wildcard_rdtypes_str}): {host}")
                else:
                    self.log.verbose(f"Finished checking {host}, it is not a wildcard")

        return wildcard_domain_results

    @property
    def dns_connectivity_lock(self):
        if self._dns_connectivity_lock is None:
            self._dns_connectivity_lock = asyncio.Lock()
        return self._dns_connectivity_lock

    async def _connectivity_check(self, interval=5):
        """
        Periodically checks for an active internet connection by attempting DNS resolution.

        Args:
            interval (int, optional): The time interval, in seconds, at which to perform the check.
            Defaults to 5 seconds.

        Returns:
            bool: True if there is an active internet connection, False otherwise.

        Examples:
            >>> await _connectivity_check()
            True
        """
        if self._last_dns_success is not None:
            if time.time() - self._last_dns_success < interval:
                return True
        dns_server_working = []
        async with self.dns_connectivity_lock:
            with suppress(Exception):
                dns_server_working = await self._catch(self.resolver.resolve, "www.google.com", rdtype="A")
                if dns_server_working:
                    self._last_dns_success = time.time()
                    return True
        if time.time() - self._last_connectivity_warning > interval:
            self.log.warning(f"DNS queries are failing, please check your internet connection")
            self._last_connectivity_warning = time.time()
        self._errors.clear()
        return False

    def debug(self, *args, **kwargs):
        if self._debug:
            self.log.trace(*args, **kwargs)

    @property
    def in_tests(self):
        return os.getenv("BBOT_TESTING", "") == "True"

    async def _mock_dns(self, mock_data):
        from .mock import MockResolver

        self.resolver = MockResolver(mock_data)



import os
import logging
from copy import copy
from pathlib import Path
from contextlib import suppress
from omegaconf import OmegaConf

from bbot.errors import BBOTError


DEFAULT_CONFIG = None


class BBOTCore:
    """
    This is the first thing that loads when you import BBOT.

    Unlike a Preset, BBOTCore holds only the config, not scan-specific stuff like targets, flags, modules, etc.

    Its main jobs are:

    - set up logging
    - keep separation between the `default` and `custom` config (this allows presets to only display the config options that have changed)
    - allow for easy merging of configs
    - load quickly
    """

    # used for filtering out sensitive config values
    secrets_strings = ["api_key", "username", "password", "token", "secret", "_id"]
    # don't filter/remove entries under this key
    secrets_exclude_keys = ["modules"]

    def __init__(self):
        self._logger = None
        self._files_config = None

        self.bbot_sudo_pass = None

        self._config = None
        self._custom_config = None

        # bare minimum == logging
        self.logger
        self.log = logging.getLogger("bbot.core")

        import multiprocessing

        self.process_name = multiprocessing.current_process().name

    @property
    def home(self):
        return Path(self.config["home"]).expanduser().resolve()

    @property
    def cache_dir(self):
        return self.home / "cache"

    @property
    def tools_dir(self):
        return self.home / "tools"

    @property
    def temp_dir(self):
        return self.home / "temp"

    @property
    def lib_dir(self):
        return self.home / "lib"

    @property
    def scans_dir(self):
        return self.home / "scans"

    @property
    def config(self):
        """
        .config is just .default_config + .custom_config merged together

        any new values should be added to custom_config.
        """
        if self._config is None:
            self._config = OmegaConf.merge(self.default_config, self.custom_config)
            # set read-only flag (change .custom_config instead)
            OmegaConf.set_readonly(self._config, True)
        return self._config

    @property
    def default_config(self):
        """
        The default BBOT config (from `defaults.yml`). Read-only.
        """
        global DEFAULT_CONFIG
        if DEFAULT_CONFIG is None:
            self.default_config = self.files_config.get_default_config()
            # ensure bbot home dir
            if not "home" in self.default_config:
                self.default_config["home"] = "~/.bbot"
        return DEFAULT_CONFIG

    @default_config.setter
    def default_config(self, value):
        # we temporarily clear out the config so it can be refreshed if/when default_config changes
        global DEFAULT_CONFIG
        self._config = None
        DEFAULT_CONFIG = value
        # set read-only flag (change .custom_config instead)
        OmegaConf.set_readonly(DEFAULT_CONFIG, True)

    @property
    def custom_config(self):
        """
        Custom BBOT config (from `~/.config/bbot/bbot.yml`)
        """
        # we temporarily clear out the config so it can be refreshed if/when custom_config changes
        self._config = None
        if self._custom_config is None:
            self.custom_config = self.files_config.get_custom_config()
        return self._custom_config

    @custom_config.setter
    def custom_config(self, value):
        # we temporarily clear out the config so it can be refreshed if/when custom_config changes
        self._config = None
        # ensure the modules key is always a dictionary
        modules_entry = value.get("modules", None)
        if modules_entry is not None and not OmegaConf.is_dict(modules_entry):
            value["modules"] = {}
        self._custom_config = value

    def no_secrets_config(self, config):
        from .helpers.misc import clean_dict

        with suppress(ValueError):
            config = OmegaConf.to_object(config)

        return clean_dict(
            config,
            *self.secrets_strings,
            fuzzy=True,
            exclude_keys=self.secrets_exclude_keys,
        )

    def secrets_only_config(self, config):
        from .helpers.misc import filter_dict

        with suppress(ValueError):
            config = OmegaConf.to_object(config)

        return filter_dict(
            config,
            *self.secrets_strings,
            fuzzy=True,
            exclude_keys=self.secrets_exclude_keys,
        )

    def merge_custom(self, config):
        """
        Merge a config into the custom config.
        """
        self.custom_config = OmegaConf.merge(self.custom_config, OmegaConf.create(config))

    def merge_default(self, config):
        """
        Merge a config into the default config.
        """
        self.default_config = OmegaConf.merge(self.default_config, OmegaConf.create(config))

    def copy(self):
        """
        Return a semi-shallow copy of self. (`custom_config` is copied, but `default_config` stays the same)
        """
        core_copy = copy(self)
        core_copy._custom_config = self._custom_config.copy()
        return core_copy

    @property
    def files_config(self):
        """
        Get the configs from `bbot.yml` and `defaults.yml`
        """
        if self._files_config is None:
            from .config import files

            self.files = files
            self._files_config = files.BBOTConfigFiles(self)
        return self._files_config

    def create_process(self, *args, **kwargs):
        if os.environ.get("BBOT_TESTING", "") == "True":
            process = self.create_thread(*args, **kwargs)
        else:
            if self.process_name == "MainProcess":
                from .helpers.process import BBOTProcess

                process = BBOTProcess(*args, **kwargs)
            else:
                raise BBOTError(f"Tried to start server from process {self.process_name}")
        process.daemon = True
        return process

    def create_thread(self, *args, **kwargs):
        from .helpers.process import BBOTThread

        return BBOTThread(*args, **kwargs)

    @property
    def logger(self):
        self.config
        if self._logger is None:
            from .config.logger import BBOTLogger

            self._logger = BBOTLogger(self)
        return self._logger



from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class TXT(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to text"}
    options = {"output_file": ""}
    options_desc = {"output_file": "Output to file"}

    output_filename = "output.txt"

    async def setup(self):
        self._prep_output_dir(self.output_filename)
        return True

    async def handle_event(self, event):
        event_str = self.human_event_str(event)

        if self.file is not None:
            self.file.write(event_str + "\n")
            self.file.flush()

    async def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    async def report(self):
        if getattr(self, "_file", None) is not None:
            self.info(f"Saved TXT output to {self.output_file}")



import json
import random
import ipaddress

from ..bbot_fixtures import *
from bbot.scanner import Scanner


@pytest.mark.asyncio
async def test_events(events, helpers):

    scan = Scanner()
    await scan._prep()

    assert events.ipv4.type == "IP_ADDRESS"
    assert events.ipv6.type == "IP_ADDRESS"
    assert events.netv4.type == "IP_RANGE"
    assert events.netv6.type == "IP_RANGE"
    assert events.domain.type == "DNS_NAME"
    assert "domain" in events.domain.tags
    assert events.subdomain.type == "DNS_NAME"
    assert "subdomain" in events.subdomain.tags
    assert events.open_port.type == "OPEN_TCP_PORT"
    assert events.url_unverified.type == "URL_UNVERIFIED"
    assert events.ipv4_url_unverified.type == "URL_UNVERIFIED"
    assert events.ipv6_url_unverified.type == "URL_UNVERIFIED"
    assert "" not in events.ipv4
    assert None not in events.ipv4
    assert 1 not in events.ipv4
    assert False not in events.ipv4

    # ip tests
    assert events.ipv4 == scan.make_event("8.8.8.8", dummy=True)
    assert "8.8.8.8" in events.ipv4
    assert "8.8.8.8" == events.ipv4
    assert "8.8.8.8" in events.netv4
    assert "8.8.8.9" not in events.ipv4
    assert "8.8.9.8" not in events.netv4
    assert "8.8.8.8/31" in events.netv4
    assert "8.8.8.8/30" in events.netv4
    assert "8.8.8.8/29" not in events.netv4
    assert "2001:4860:4860::8888" in events.ipv6
    assert "2001:4860:4860::8888" in events.netv6
    assert "2001:4860:4860::8889" not in events.ipv6
    assert "2002:4860:4860::8888" not in events.netv6
    assert "2001:4860:4860::8888/127" in events.netv6
    assert "2001:4860:4860::8888/126" in events.netv6
    assert "2001:4860:4860::8888/125" not in events.netv6
    assert events.emoji not in events.ipv4
    assert events.emoji not in events.netv6
    assert events.netv6 not in events.emoji
    assert "dead::c0de" == scan.make_event(" [DEaD::c0De]:88", "DNS_NAME", dummy=True)

    # hostname tests
    assert events.domain.host == "publicapis.org"
    assert events.subdomain.host == "api.publicapis.org"
    assert events.domain.host_stem == "publicapis"
    assert events.subdomain.host_stem == "api.publicapis"
    assert "api.publicapis.org" in events.domain
    assert "api.publicapis.org" in events.subdomain
    assert "fsocie.ty" not in events.domain
    assert "fsocie.ty" not in events.subdomain
    assert events.subdomain in events.domain
    assert events.domain not in events.subdomain
    assert not events.ipv4 in events.domain
    assert not events.netv6 in events.domain
    assert events.emoji not in events.domain
    assert events.domain not in events.emoji
    assert "evilcorp.com" == scan.make_event(" eViLcorp.COM.:88", "DNS_NAME", dummy=True)
    assert "evilcorp.com" == scan.make_event("evilcorp.com.", "DNS_NAME", dummy=True)

    # url tests
    assert scan.make_event("http://evilcorp.com", dummy=True) == scan.make_event("http://evilcorp.com/", dummy=True)
    assert events.url_unverified.host == "api.publicapis.org"
    assert events.url_unverified in events.domain
    assert events.url_unverified in events.subdomain
    assert "api.publicapis.org:443" in events.url_unverified
    assert "publicapis.org" not in events.url_unverified
    assert events.ipv4_url_unverified in events.ipv4
    assert events.ipv4_url_unverified in events.netv4
    assert events.ipv6_url_unverified in events.ipv6
    assert events.ipv6_url_unverified in events.netv6
    assert events.emoji not in events.url_unverified
    assert events.emoji not in events.ipv6_url_unverified
    assert events.url_unverified not in events.emoji
    assert "https://evilcorp.com" == scan.make_event("https://evilcorp.com:443", dummy=True)
    assert "http://evilcorp.com" == scan.make_event("http://evilcorp.com:80", dummy=True)
    assert "http://evilcorp.com:80/asdf.js" in scan.make_event("http://evilcorp.com/asdf.js", dummy=True)
    assert "http://evilcorp.com/asdf.js" in scan.make_event("http://evilcorp.com:80/asdf.js", dummy=True)
    assert "https://evilcorp.com:443" == scan.make_event("https://evilcorp.com", dummy=True)
    assert "http://evilcorp.com:80" == scan.make_event("http://evilcorp.com", dummy=True)
    assert "https://evilcorp.com:80" == scan.make_event("https://evilcorp.com:80", dummy=True)
    assert "http://evilcorp.com:443" == scan.make_event("http://evilcorp.com:443", dummy=True)
    assert scan.make_event("https://evilcorp.com", dummy=True).with_port().geturl() == "https://evilcorp.com:443/"
    assert scan.make_event("https://evilcorp.com:666", dummy=True).with_port().geturl() == "https://evilcorp.com:666/"
    assert scan.make_event("https://evilcorp.com.:666", dummy=True) == "https://evilcorp.com:666/"
    assert scan.make_event("https://[bad::c0de]", dummy=True).with_port().geturl() == "https://[bad::c0de]:443/"
    assert scan.make_event("https://[bad::c0de]:666", dummy=True).with_port().geturl() == "https://[bad::c0de]:666/"
    url_event = scan.make_event("https://evilcorp.com", "URL", events.ipv4_url, tags=["status-200"])
    assert "status-200" in url_event.tags
    assert url_event.http_status == 200
    with pytest.raises(ValidationError, match=".*status tag.*"):
        scan.make_event("https://evilcorp.com", "URL", events.ipv4_url)

    # http response
    assert events.http_response.host == "example.com"
    assert events.http_response.port == 80
    assert events.http_response.parsed_url.scheme == "http"
    assert events.http_response.with_port().geturl() == "http://example.com:80/"

    http_response = scan.make_event(
        {
            "port": "80",
            "title": "HTTP%20RESPONSE",
            "url": "http://www.evilcorp.com:80",
            "input": "http://www.evilcorp.com:80",
            "location": "/asdf",
            "status_code": 301,
        },
        "HTTP_RESPONSE",
        dummy=True,
    )
    assert http_response.http_status == 301
    assert http_response.http_title == "HTTP RESPONSE"
    assert http_response.redirect_location == "http://www.evilcorp.com/asdf"

    # http response url validation
    http_response_2 = scan.make_event(
        {"port": "80", "url": "http://evilcorp.com:80/asdf"}, "HTTP_RESPONSE", dummy=True
    )
    assert http_response_2.data["url"] == "http://evilcorp.com/asdf"

    # open port tests
    assert events.open_port in events.domain
    assert "api.publicapis.org:443" in events.open_port
    assert "bad.publicapis.org:443" not in events.open_port
    assert "publicapis.org:443" not in events.open_port
    assert events.ipv4_open_port in events.ipv4
    assert events.ipv4_open_port in events.netv4
    assert "8.8.8.9" not in events.ipv4_open_port
    assert events.ipv6_open_port in events.ipv6
    assert events.ipv6_open_port in events.netv6
    assert "2002:4860:4860::8888" not in events.ipv6_open_port
    assert events.emoji not in events.ipv6_open_port
    assert events.ipv6_open_port not in events.emoji

    # attribute tests
    assert events.ipv4.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4.port is None
    assert events.ipv6.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6.port is None
    assert events.domain.port is None
    assert events.subdomain.port is None
    assert events.open_port.host == "api.publicapis.org"
    assert events.open_port.port == 443
    assert events.ipv4_open_port.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4_open_port.port == 443
    assert events.ipv6_open_port.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6_open_port.port == 443
    assert events.url_unverified.host == "api.publicapis.org"
    assert events.url_unverified.port == 443
    assert events.ipv4_url_unverified.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4_url_unverified.port == 443
    assert events.ipv6_url_unverified.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6_url_unverified.port == 443

    javascript_event = scan.make_event("http://evilcorp.com/asdf/a.js?b=c#d", "URL_UNVERIFIED", parent=scan.root_event)
    assert "extension-js" in javascript_event.tags
    await scan.ingress_module.handle_event(javascript_event)
    assert "httpx-only" in javascript_event.tags

    # scope distance
    event1 = scan.make_event("1.2.3.4", dummy=True)
    assert event1._scope_distance == None
    event1.scope_distance = 0
    assert event1._scope_distance == 0
    event2 = scan.make_event("2.3.4.5", parent=event1)
    assert event2._scope_distance == 1
    event3 = scan.make_event("3.4.5.6", parent=event2)
    assert event3._scope_distance == 2
    event4 = scan.make_event("3.4.5.6", parent=event3)
    assert event4._scope_distance == 2
    event5 = scan.make_event("4.5.6.7", parent=event4)
    assert event5._scope_distance == 3

    url_1 = scan.make_event("https://127.0.0.1/asdf", "URL_UNVERIFIED", parent=scan.root_event)
    assert url_1.scope_distance == 1
    url_2 = scan.make_event("https://127.0.0.1/test", "URL_UNVERIFIED", parent=url_1)
    assert url_2.scope_distance == 1
    url_3 = scan.make_event("https://127.0.0.2/asdf", "URL_UNVERIFIED", parent=url_1)
    assert url_3.scope_distance == 2

    org_stub_1 = scan.make_event("STUB1", "ORG_STUB", parent=scan.root_event)
    org_stub_1.scope_distance == 1
    org_stub_2 = scan.make_event("STUB2", "ORG_STUB", parent=org_stub_1)
    org_stub_2.scope_distance == 2

    # internal event tracking
    root_event = scan.make_event("0.0.0.0", dummy=True)
    root_event.scope_distance = 0
    internal_event1 = scan.make_event("1.2.3.4", parent=root_event, internal=True)
    assert internal_event1._internal == True
    assert "internal" in internal_event1.tags

    # tag inheritance
    for tag in ("affiliate", "mutation-1"):
        affiliate_event = scan.make_event("1.2.3.4", parent=root_event, tags=tag)
        assert tag in affiliate_event.tags
        affiliate_event2 = scan.make_event("1.2.3.4:88", parent=affiliate_event)
        affiliate_event3 = scan.make_event("4.3.2.1:88", parent=affiliate_event)
        assert tag in affiliate_event2.tags
        assert tag not in affiliate_event3.tags

    # discovery context
    event = scan.make_event(
        "127.0.0.1", parent=scan.root_event, context="something discovered {event.type}: {event.data}"
    )
    assert event.discovery_context == "something discovered IP_ADDRESS: 127.0.0.1"

    # updating an already-created event with make_event()
    # updating tags
    event1 = scan.make_event("127.0.0.1", parent=scan.root_event)
    updated_event = scan.make_event(event1, tags="asdf")
    assert "asdf" not in event1.tags
    assert "asdf" in updated_event.tags
    # updating parent
    event2 = scan.make_event("127.0.0.1", parent=scan.root_event)
    updated_event = scan.make_event(event2, parent=event1)
    assert event2.parent == scan.root_event
    assert updated_event.parent == event1
    # updating module
    event3 = scan.make_event("127.0.0.1", parent=scan.root_event)
    updated_event = scan.make_event(event3, internal=True)
    assert event3.internal == False
    assert updated_event.internal == True

    # event sorting
    parent1 = scan.make_event("127.0.0.1", parent=scan.root_event)
    parent2 = scan.make_event("127.0.0.1", parent=scan.root_event)
    parent2_child1 = scan.make_event("127.0.0.1", parent=parent2)
    parent1_child1 = scan.make_event("127.0.0.1", parent=parent1)
    parent1_child2 = scan.make_event("127.0.0.1", parent=parent1)
    parent1_child2_child1 = scan.make_event("127.0.0.1", parent=parent1_child2)
    parent1_child2_child2 = scan.make_event("127.0.0.1", parent=parent1_child2)
    parent1_child1_child1 = scan.make_event("127.0.0.1", parent=parent1_child1)
    parent2_child2 = scan.make_event("127.0.0.1", parent=parent2)
    parent1_child2_child1_child1 = scan.make_event("127.0.0.1", parent=parent1_child2_child1)

    sortable_events = {
        "parent1": parent1,
        "parent2": parent2,
        "parent2_child1": parent2_child1,
        "parent1_child1": parent1_child1,
        "parent1_child2": parent1_child2,
        "parent1_child2_child1": parent1_child2_child1,
        "parent1_child2_child2": parent1_child2_child2,
        "parent1_child1_child1": parent1_child1_child1,
        "parent2_child2": parent2_child2,
        "parent1_child2_child1_child1": parent1_child2_child1_child1,
    }

    ordered_list = [
        parent1,
        parent1_child1,
        parent1_child1_child1,
        parent1_child2,
        parent1_child2_child1,
        parent1_child2_child1_child1,
        parent1_child2_child2,
        parent2,
        parent2_child1,
        parent2_child2,
    ]

    shuffled_list = list(sortable_events.values())
    random.shuffle(shuffled_list)

    sorted_events = sorted(shuffled_list)
    assert sorted_events == ordered_list

    # test validation
    corrected_event1 = scan.make_event("asdf@asdf.com", "DNS_NAME", dummy=True)
    assert corrected_event1.type == "EMAIL_ADDRESS"
    corrected_event2 = scan.make_event("127.0.0.1", "DNS_NAME", dummy=True)
    assert corrected_event2.type == "IP_ADDRESS"
    corrected_event3 = scan.make_event("wat.asdf.com", "IP_ADDRESS", dummy=True)
    assert corrected_event3.type == "DNS_NAME"

    corrected_event4 = scan.make_event("bob@evilcorp.com", "USERNAME", dummy=True)
    assert corrected_event4.type == "EMAIL_ADDRESS"
    assert "affiliate" in corrected_event4.tags

    test_vuln = scan.make_event(
        {"host": "EVILcorp.com", "severity": "iNfo ", "description": "asdf"}, "VULNERABILITY", dummy=True
    )
    assert test_vuln.data["host"] == "evilcorp.com"
    assert test_vuln.data["severity"] == "INFO"
    test_vuln2 = scan.make_event(
        {"host": "192.168.1.1", "severity": "iNfo ", "description": "asdf"}, "VULNERABILITY", dummy=True
    )
    assert json.loads(test_vuln2.data_human)["severity"] == "INFO"
    assert test_vuln2.host.is_private
    with pytest.raises(ValidationError, match=".*validation error.*\nseverity\n.*Field required.*"):
        test_vuln = scan.make_event({"host": "evilcorp.com", "description": "asdf"}, "VULNERABILITY", dummy=True)
    with pytest.raises(ValidationError, match=".*host.*\n.*Invalid host.*"):
        test_vuln = scan.make_event(
            {"host": "!@#$", "severity": "INFO", "description": "asdf"}, "VULNERABILITY", dummy=True
        )
    with pytest.raises(ValidationError, match=".*severity.*\n.*Invalid severity.*"):
        test_vuln = scan.make_event(
            {"host": "evilcorp.com", "severity": "WACK", "description": "asdf"}, "VULNERABILITY", dummy=True
        )

    # punycode - event type detection

    # japanese
    assert scan.make_event("ドメイン.テスト", dummy=True).type == "DNS_NAME"
    assert scan.make_event("bob@ドメイン.テスト", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("テスト@ドメイン.テスト", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("ドメイン.テスト:80", dummy=True).type == "OPEN_TCP_PORT"
    assert scan.make_event("http://ドメイン.テスト:80", dummy=True).type == "URL_UNVERIFIED"
    assert scan.make_event("http://ドメイン.テスト:80/テスト", dummy=True).type == "URL_UNVERIFIED"

    assert scan.make_event("xn--eckwd4c7c.xn--zckzah", dummy=True).type == "DNS_NAME"
    assert scan.make_event("bob@xn--eckwd4c7c.xn--zckzah", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("テスト@xn--eckwd4c7c.xn--zckzah", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("xn--eckwd4c7c.xn--zckzah:80", dummy=True).type == "OPEN_TCP_PORT"
    assert scan.make_event("http://xn--eckwd4c7c.xn--zckzah:80", dummy=True).type == "URL_UNVERIFIED"
    assert scan.make_event("http://xn--eckwd4c7c.xn--zckzah:80/テスト", dummy=True).type == "URL_UNVERIFIED"

    # thai
    assert scan.make_event("เราเที่ยวด้วยกัน.com", dummy=True).type == "DNS_NAME"
    assert scan.make_event("bob@เราเที่ยวด้วยกัน.com", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("ทดสอบ@เราเที่ยวด้วยกัน.com", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("เราเที่ยวด้วยกัน.com:80", dummy=True).type == "OPEN_TCP_PORT"
    assert scan.make_event("http://เราเที่ยวด้วยกัน.com:80", dummy=True).type == "URL_UNVERIFIED"
    assert scan.make_event("http://เราเที่ยวด้วยกัน.com:80/ทดสอบ", dummy=True).type == "URL_UNVERIFIED"

    assert scan.make_event("xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).type == "DNS_NAME"
    assert scan.make_event("bob@xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("ทดสอบ@xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).type == "EMAIL_ADDRESS"
    assert scan.make_event("xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80", dummy=True).type == "OPEN_TCP_PORT"
    assert scan.make_event("http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80", dummy=True).type == "URL_UNVERIFIED"
    assert scan.make_event("http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80/ทดสอบ", dummy=True).type == "URL_UNVERIFIED"

    # punycode - encoding / decoding tests

    # japanese
    assert scan.make_event("xn--eckwd4c7c.xn--zckzah", dummy=True).data == "xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("bob@xn--eckwd4c7c.xn--zckzah", dummy=True).data == "bob@xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("テスト@xn--eckwd4c7c.xn--zckzah", dummy=True).data == "テスト@xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("xn--eckwd4c7c.xn--zckzah:80", dummy=True).data == "xn--eckwd4c7c.xn--zckzah:80"
    assert scan.make_event("http://xn--eckwd4c7c.xn--zckzah:80", dummy=True).data == "http://xn--eckwd4c7c.xn--zckzah/"
    assert (
        scan.make_event("http://xn--eckwd4c7c.xn--zckzah:80/テスト", dummy=True).data
        == "http://xn--eckwd4c7c.xn--zckzah/テスト"
    )

    assert scan.make_event("ドメイン.テスト", dummy=True).data == "xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("bob@ドメイン.テスト", dummy=True).data == "bob@xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("テスト@ドメイン.テスト", dummy=True).data == "テスト@xn--eckwd4c7c.xn--zckzah"
    assert scan.make_event("ドメイン.テスト:80", dummy=True).data == "xn--eckwd4c7c.xn--zckzah:80"
    assert scan.make_event("http://ドメイン.テスト:80", dummy=True).data == "http://xn--eckwd4c7c.xn--zckzah/"
    assert (
        scan.make_event("http://ドメイン.テスト:80/テスト", dummy=True).data
        == "http://xn--eckwd4c7c.xn--zckzah/テスト"
    )
    # thai
    assert (
        scan.make_event("xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).data == "xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    )
    assert (
        scan.make_event("bob@xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).data
        == "bob@xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    )
    assert (
        scan.make_event("ทดสอบ@xn--12c1bik6bbd8ab6hd1b5jc6jta.com", dummy=True).data
        == "ทดสอบ@xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    )
    assert (
        scan.make_event("xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80", dummy=True).data
        == "xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80"
    )
    assert (
        scan.make_event("http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80", dummy=True).data
        == "http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com/"
    )
    assert (
        scan.make_event("http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80/ทดสอบ", dummy=True).data
        == "http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com/ทดสอบ"
    )

    assert scan.make_event("เราเที่ยวด้วยกัน.com", dummy=True).data == "xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    assert scan.make_event("bob@เราเที่ยวด้วยกัน.com", dummy=True).data == "bob@xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    assert scan.make_event("ทดสอบ@เราเที่ยวด้วยกัน.com", dummy=True).data == "ทดสอบ@xn--12c1bik6bbd8ab6hd1b5jc6jta.com"
    assert scan.make_event("เราเที่ยวด้วยกัน.com:80", dummy=True).data == "xn--12c1bik6bbd8ab6hd1b5jc6jta.com:80"
    assert (
        scan.make_event("http://เราเที่ยวด้วยกัน.com:80", dummy=True).data == "http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com/"
    )
    assert (
        scan.make_event("http://เราเที่ยวด้วยกัน.com:80/ทดสอบ", dummy=True).data
        == "http://xn--12c1bik6bbd8ab6hd1b5jc6jta.com/ทดสอบ"
    )

    # test event serialization
    from bbot.core.event import event_from_json

    db_event = scan.make_event("evilcorp.com:80", parent=scan.root_event, context="test context")
    db_event._resolved_hosts = {"127.0.0.1"}
    db_event.scope_distance = 1
    assert db_event.discovery_context == "test context"
    assert db_event.discovery_path == ["test context"]
    assert db_event.parent_chain == ["OPEN_TCP_PORT:5098b5e3fc65b13bb4a5cee4201c2e160fa4ffac"]
    timestamp = db_event.timestamp.isoformat()
    json_event = db_event.json()
    assert json_event["scope_distance"] == 1
    assert json_event["data"] == "evilcorp.com:80"
    assert json_event["type"] == "OPEN_TCP_PORT"
    assert json_event["host"] == "evilcorp.com"
    assert json_event["timestamp"] == timestamp
    assert json_event["discovery_context"] == "test context"
    assert json_event["discovery_path"] == ["test context"]
    assert json_event["parent_chain"] == ["OPEN_TCP_PORT:5098b5e3fc65b13bb4a5cee4201c2e160fa4ffac"]
    reconstituted_event = event_from_json(json_event)
    assert reconstituted_event.scope_distance == 1
    assert reconstituted_event.timestamp.isoformat() == timestamp
    assert reconstituted_event.data == "evilcorp.com:80"
    assert reconstituted_event.type == "OPEN_TCP_PORT"
    assert reconstituted_event.host == "evilcorp.com"
    assert reconstituted_event.discovery_context == "test context"
    assert reconstituted_event.discovery_path == ["test context"]
    assert reconstituted_event.parent_chain == ["OPEN_TCP_PORT:5098b5e3fc65b13bb4a5cee4201c2e160fa4ffac"]
    assert "127.0.0.1" in reconstituted_event.resolved_hosts
    hostless_event = scan.make_event("asdf", "ASDF", dummy=True)
    hostless_event_json = hostless_event.json()
    assert hostless_event_json["type"] == "ASDF"
    assert hostless_event_json["data"] == "asdf"
    assert not "host" in hostless_event_json

    # SIEM-friendly serialize/deserialize
    json_event_siemfriendly = db_event.json(siem_friendly=True)
    assert json_event_siemfriendly["scope_distance"] == 1
    assert json_event_siemfriendly["data"] == {"OPEN_TCP_PORT": "evilcorp.com:80"}
    assert json_event_siemfriendly["type"] == "OPEN_TCP_PORT"
    assert json_event_siemfriendly["host"] == "evilcorp.com"
    assert json_event_siemfriendly["timestamp"] == timestamp
    reconstituted_event2 = event_from_json(json_event_siemfriendly, siem_friendly=True)
    assert reconstituted_event2.scope_distance == 1
    assert reconstituted_event2.timestamp.isoformat() == timestamp
    assert reconstituted_event2.data == "evilcorp.com:80"
    assert reconstituted_event2.type == "OPEN_TCP_PORT"
    assert reconstituted_event2.host == "evilcorp.com"
    assert "127.0.0.1" in reconstituted_event2.resolved_hosts

    http_response = scan.make_event(httpx_response, "HTTP_RESPONSE", parent=scan.root_event)
    assert http_response.parent_id == scan.root_event.id
    assert http_response.data["input"] == "http://example.com:80"
    json_event = http_response.json(mode="graph")
    assert isinstance(json_event["data"], str)
    json_event = http_response.json()
    assert isinstance(json_event["data"], dict)
    assert json_event["type"] == "HTTP_RESPONSE"
    assert json_event["host"] == "example.com"
    assert json_event["parent"] == scan.root_event.id
    reconstituted_event = event_from_json(json_event)
    assert isinstance(reconstituted_event.data, dict)
    assert reconstituted_event.data["input"] == "http://example.com:80"
    assert reconstituted_event.host == "example.com"
    assert reconstituted_event.type == "HTTP_RESPONSE"
    assert reconstituted_event.parent_id == scan.root_event.id

    event_1 = scan.make_event("127.0.0.1", parent=scan.root_event)
    event_2 = scan.make_event("127.0.0.2", parent=event_1)
    event_3 = scan.make_event("127.0.0.3", parent=event_2)
    event_3._omit = True
    event_4 = scan.make_event("127.0.0.4", parent=event_3)
    event_5 = scan.make_event("127.0.0.5", parent=event_4)
    assert event_5.get_parents() == [event_4, event_3, event_2, event_1, scan.root_event]
    assert event_5.get_parents(omit=True) == [event_4, event_2, event_1, scan.root_event]
    assert event_5.get_parents(include_self=True) == [event_5, event_4, event_3, event_2, event_1, scan.root_event]

    # test host backup
    host_event = scan.make_event("asdf.evilcorp.com", "DNS_NAME", parent=scan.root_event)
    assert host_event.host_original == "asdf.evilcorp.com"
    host_event.host = "_wildcard.evilcorp.com"
    assert host_event.host == "_wildcard.evilcorp.com"
    assert host_event.host_original == "asdf.evilcorp.com"

    # test storage bucket validation
    bucket_event = scan.make_event(
        {"name": "ASDF.s3.amazonaws.com", "url": "https://ASDF.s3.amazonaws.com"},
        "STORAGE_BUCKET",
        parent=scan.root_event,
    )
    assert bucket_event.data["name"] == "asdf.s3.amazonaws.com"
    assert bucket_event.data["url"] == "https://asdf.s3.amazonaws.com/"

    # test module sequence
    module = scan._make_dummy_module("mymodule")
    parent_event_1 = scan.make_event("127.0.0.1", module=module, parent=scan.root_event)
    assert str(parent_event_1.module) == "mymodule"
    assert str(parent_event_1.module_sequence) == "mymodule"
    parent_event_2 = scan.make_event("127.0.0.2", module=module, parent=parent_event_1)
    assert str(parent_event_2.module) == "mymodule"
    assert str(parent_event_2.module_sequence) == "mymodule"
    parent_event_3 = scan.make_event("127.0.0.3", module=module, parent=parent_event_2)
    assert str(parent_event_3.module) == "mymodule"
    assert str(parent_event_3.module_sequence) == "mymodule"

    module = scan._make_dummy_module("mymodule")
    parent_event_1 = scan.make_event("127.0.0.1", module=module, parent=scan.root_event)
    parent_event_1._omit = True
    assert str(parent_event_1.module) == "mymodule"
    assert str(parent_event_1.module_sequence) == "mymodule"
    parent_event_2 = scan.make_event("127.0.0.2", module=module, parent=parent_event_1)
    parent_event_2._omit = True
    assert str(parent_event_2.module) == "mymodule"
    assert str(parent_event_2.module_sequence) == "mymodule->mymodule"
    parent_event_3 = scan.make_event("127.0.0.3", module=module, parent=parent_event_2)
    assert str(parent_event_3.module) == "mymodule"
    assert str(parent_event_3.module_sequence) == "mymodule->mymodule->mymodule"

    await scan._cleanup()


@pytest.mark.asyncio
async def test_event_discovery_context():

    from bbot.modules.base import BaseModule

    scan = Scanner("evilcorp.com")
    await scan.helpers.dns._mock_dns(
        {
            "evilcorp.com": {"A": ["1.2.3.4"]},
            "one.evilcorp.com": {"A": ["1.2.3.4"]},
            "two.evilcorp.com": {"A": ["1.2.3.4"]},
            "three.evilcorp.com": {"A": ["1.2.3.4"]},
            "four.evilcorp.com": {"A": ["1.2.3.4"]},
        }
    )
    await scan._prep()

    dummy_module_1 = scan._make_dummy_module("module_1")
    dummy_module_2 = scan._make_dummy_module("module_2")

    class DummyModule(BaseModule):
        watched_events = ["DNS_NAME"]
        _name = "dummy_module"

        async def handle_event(self, event):
            new_event = None
            if event.data == "evilcorp.com":
                new_event = scan.make_event(
                    "one.evilcorp.com",
                    "DNS_NAME",
                    event,
                    context="{module} invoked forbidden magick to discover {event.type} {event.data}",
                    module=dummy_module_1,
                )
            elif event.data == "one.evilcorp.com":
                new_event = scan.make_event(
                    "two.evilcorp.com",
                    "DNS_NAME",
                    event,
                    context="{module} pledged its allegiance to cthulu and was awarded {event.type} {event.data}",
                    module=dummy_module_1,
                )
            elif event.data == "two.evilcorp.com":
                new_event = scan.make_event(
                    "three.evilcorp.com",
                    "DNS_NAME",
                    event,
                    context="{module} asked nicely and was given {event.type} {event.data}",
                    module=dummy_module_2,
                )
            elif event.data == "three.evilcorp.com":
                new_event = scan.make_event(
                    "four.evilcorp.com",
                    "DNS_NAME",
                    event,
                    context="{module} used brute force to obtain {event.type} {event.data}",
                    module=dummy_module_2,
                )
            if new_event is not None:
                await self.emit_event(new_event)

    dummy_module = DummyModule(scan)

    scan.modules["dummy_module"] = dummy_module

    # test discovery context
    test_event = dummy_module.make_event("evilcorp.com", "DNS_NAME", parent=scan.root_event)
    assert test_event.discovery_context == "dummy_module discovered DNS_NAME: evilcorp.com"

    test_event2 = dummy_module.make_event(
        "evilcorp.com", "DNS_NAME", parent=scan.root_event, context="{module} {found} {event.host}"
    )
    assert test_event2.discovery_context == "dummy_module {found} evilcorp.com"
    # jank input
    test_event3 = dummy_module.make_event(
        "http://evilcorp.com/{http://evilcorp.org!@#%@#$:,,,}", "URL_UNVERIFIED", parent=scan.root_event
    )
    assert (
        test_event3.discovery_context
        == "dummy_module discovered URL_UNVERIFIED: http://evilcorp.com/{http:/evilcorp.org!@"
    )

    events = [e async for e in scan.async_start()]
    assert len(events) == 6

    assert 1 == len(
        [
            e
            for e in events
            if e.type == "DNS_NAME"
            and e.data == "evilcorp.com"
            and e.discovery_context == f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com"
            and e.discovery_path == [f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com"]
        ]
    )
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "DNS_NAME"
            and e.data == "one.evilcorp.com"
            and e.discovery_context == "module_1 invoked forbidden magick to discover DNS_NAME one.evilcorp.com"
            and e.discovery_path
            == [
                f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com",
                "module_1 invoked forbidden magick to discover DNS_NAME one.evilcorp.com",
            ]
        ]
    )
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "DNS_NAME"
            and e.data == "two.evilcorp.com"
            and e.discovery_context
            == "module_1 pledged its allegiance to cthulu and was awarded DNS_NAME two.evilcorp.com"
            and e.discovery_path
            == [
                f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com",
                "module_1 invoked forbidden magick to discover DNS_NAME one.evilcorp.com",
                "module_1 pledged its allegiance to cthulu and was awarded DNS_NAME two.evilcorp.com",
            ]
        ]
    )
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "DNS_NAME"
            and e.data == "three.evilcorp.com"
            and e.discovery_context == "module_2 asked nicely and was given DNS_NAME three.evilcorp.com"
            and e.discovery_path
            == [
                f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com",
                "module_1 invoked forbidden magick to discover DNS_NAME one.evilcorp.com",
                "module_1 pledged its allegiance to cthulu and was awarded DNS_NAME two.evilcorp.com",
                "module_2 asked nicely and was given DNS_NAME three.evilcorp.com",
            ]
        ]
    )
    final_path = [
        f"Scan {scan.name} seeded with DNS_NAME: evilcorp.com",
        "module_1 invoked forbidden magick to discover DNS_NAME one.evilcorp.com",
        "module_1 pledged its allegiance to cthulu and was awarded DNS_NAME two.evilcorp.com",
        "module_2 asked nicely and was given DNS_NAME three.evilcorp.com",
        "module_2 used brute force to obtain DNS_NAME four.evilcorp.com",
    ]
    final_event = [
        e
        for e in events
        if e.type == "DNS_NAME"
        and e.data == "four.evilcorp.com"
        and e.discovery_context == "module_2 used brute force to obtain DNS_NAME four.evilcorp.com"
        and e.discovery_path == final_path
    ]
    assert 1 == len(final_event)
    j = final_event[0].json()
    assert j["discovery_path"] == final_path

    await scan._cleanup()

    # test to make sure this doesn't come back
    #  https://github.com/blacklanternsecurity/bbot/issues/1498
    scan = Scanner("http://blacklanternsecurity.com", config={"dns": {"minimal": False}})
    await scan.helpers.dns._mock_dns(
        {"blacklanternsecurity.com": {"TXT": ["blsops.com"], "A": ["127.0.0.1"]}, "blsops.com": {"A": ["127.0.0.1"]}}
    )
    events = [e async for e in scan.async_start()]
    blsops_event = [e for e in events if e.type == "DNS_NAME" and e.data == "blsops.com"]
    assert len(blsops_event) == 1
    assert blsops_event[0].discovery_path[1] == "URL_UNVERIFIED has host DNS_NAME: blacklanternsecurity.com"

    await scan._cleanup()


@pytest.mark.asyncio
async def test_event_web_spider_distance(bbot_scanner):
    # make sure web spider distance inheritance works as intended
    # and we don't have any runaway situations with SOCIAL events + URLs
    scan = bbot_scanner(config={"web": {"spider_distance": 1}})
    url_event_1 = scan.make_event("http://www.evilcorp.com/test1", "URL_UNVERIFIED", parent=scan.root_event)
    assert url_event_1.web_spider_distance == 0
    url_event_2 = scan.make_event("http://www.evilcorp.com/test2", "URL_UNVERIFIED", parent=url_event_1)
    assert url_event_2.web_spider_distance == 0
    url_event_3 = scan.make_event(
        "http://www.evilcorp.com/test3", "URL_UNVERIFIED", parent=url_event_2, tags=["spider-danger"]
    )
    assert url_event_3.web_spider_distance == 1
    assert "spider-danger" in url_event_3.tags
    assert not "spider-max" in url_event_3.tags
    social_event = scan.make_event(
        {"platform": "github", "url": "http://www.evilcorp.com/test4"}, "SOCIAL", parent=url_event_3
    )
    assert social_event.web_spider_distance == 1
    assert "spider-danger" in social_event.tags
    url_event_4 = scan.make_event("http://www.evilcorp.com/test4", "URL_UNVERIFIED", parent=social_event)
    assert url_event_4.web_spider_distance == 2
    assert "spider-danger" in url_event_4.tags
    assert "spider-max" in url_event_4.tags
    social_event_2 = scan.make_event(
        {"platform": "github", "url": "http://www.evilcorp.com/test5"}, "SOCIAL", parent=url_event_4
    )
    assert social_event_2.web_spider_distance == 2
    assert "spider-danger" in social_event_2.tags
    assert "spider-max" in social_event_2.tags
    url_event_5 = scan.make_event("http://www.evilcorp.com/test5", "URL_UNVERIFIED", parent=social_event_2)
    assert url_event_5.web_spider_distance == 3
    assert "spider-danger" in url_event_5.tags
    assert "spider-max" in url_event_5.tags

    url_event = scan.make_event("http://www.evilcorp.com", "URL_UNVERIFIED", parent=scan.root_event)
    assert url_event.web_spider_distance == 0
    assert not "spider-danger" in url_event.tags
    assert not "spider-max" in url_event.tags
    url_event_2 = scan.make_event(
        "http://www.evilcorp.com", "URL_UNVERIFIED", parent=scan.root_event, tags="spider-danger"
    )
    # spider distance shouldn't increment because it's not the same host
    assert url_event_2.web_spider_distance == 0
    assert "spider-danger" in url_event_2.tags
    assert not "spider-max" in url_event_2.tags
    url_event_3 = scan.make_event(
        "http://www.evilcorp.com/3", "URL_UNVERIFIED", parent=url_event_2, tags="spider-danger"
    )
    assert url_event_3.web_spider_distance == 1
    assert "spider-danger" in url_event_3.tags
    assert not "spider-max" in url_event_3.tags
    url_event_4 = scan.make_event("http://evilcorp.com", "URL_UNVERIFIED", parent=url_event_3)
    assert url_event_4.web_spider_distance == 0
    assert not "spider-danger" in url_event_4.tags
    assert not "spider-max" in url_event_4.tags
    url_event_4.add_tag("spider-danger")
    assert url_event_4.web_spider_distance == 0
    assert "spider-danger" in url_event_4.tags
    assert not "spider-max" in url_event_4.tags
    url_event_4.remove_tag("spider-danger")
    assert url_event_4.web_spider_distance == 0
    assert not "spider-danger" in url_event_4.tags
    assert not "spider-max" in url_event_4.tags
    url_event_5 = scan.make_event("http://evilcorp.com/5", "URL_UNVERIFIED", parent=url_event_4)
    assert url_event_5.web_spider_distance == 0
    assert not "spider-danger" in url_event_5.tags
    assert not "spider-max" in url_event_5.tags
    url_event_5.add_tag("spider-danger")
    # if host is the same as parent, web spider distance should auto-increment after adding spider-danger tag
    assert url_event_5.web_spider_distance == 1
    assert "spider-danger" in url_event_5.tags
    assert not "spider-max" in url_event_5.tags


def test_event_confidence():
    scan = Scanner()
    # default 100
    event1 = scan.make_event("evilcorp.com", "DNS_NAME", dummy=True)
    assert event1.confidence == 100
    assert event1.cumulative_confidence == 100
    # custom confidence
    event2 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=90, dummy=True)
    assert event2.confidence == 90
    assert event2.cumulative_confidence == 90
    # max 100
    event3 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=999, dummy=True)
    assert event3.confidence == 100
    assert event3.cumulative_confidence == 100
    # min 1
    event4 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=0, dummy=True)
    assert event4.confidence == 1
    assert event4.cumulative_confidence == 1
    # first event in chain
    event5 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=90, parent=scan.root_event)
    assert event5.confidence == 90
    assert event5.cumulative_confidence == 90
    # compounding confidence
    event6 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=50, parent=event5)
    assert event6.confidence == 50
    assert event6.cumulative_confidence == 45
    event7 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=50, parent=event6)
    assert event7.confidence == 50
    assert event7.cumulative_confidence == 22
    # 100 confidence resets
    event8 = scan.make_event("evilcorp.com", "DNS_NAME", confidence=100, parent=event7)
    assert event8.confidence == 100
    assert event8.cumulative_confidence == 100


def test_event_closest_host():
    scan = Scanner()
    # first event has a host
    event1 = scan.make_event("evilcorp.com", "DNS_NAME", parent=scan.root_event)
    assert event1.host == "evilcorp.com"
    # second event has a host + url
    event2 = scan.make_event(
        {"method": "GET", "url": "http://www.evilcorp.com/asdf", "hash": {"header_mmh3": "1", "body_mmh3": "2"}},
        "HTTP_RESPONSE",
        parent=event1,
    )
    assert event2.host == "www.evilcorp.com"
    # third event has a path
    event3 = scan.make_event({"path": "/tmp/asdf.txt"}, "FILESYSTEM", parent=event2)
    assert not event3.host
    # finding automatically uses the host from the second event
    finding = scan.make_event({"description": "test"}, "FINDING", parent=event3)
    assert finding.data["host"] == "www.evilcorp.com"
    assert finding.data["url"] == "http://www.evilcorp.com/asdf"
    assert finding.data["path"] == "/tmp/asdf.txt"
    assert finding.host == "www.evilcorp.com"
    # same with vuln
    vuln = scan.make_event({"description": "test", "severity": "HIGH"}, "VULNERABILITY", parent=event3)
    assert vuln.data["host"] == "www.evilcorp.com"
    assert vuln.data["url"] == "http://www.evilcorp.com/asdf"
    assert vuln.data["path"] == "/tmp/asdf.txt"
    assert vuln.host == "www.evilcorp.com"

    # no host == not allowed
    event3 = scan.make_event("wat", "ASDF", parent=scan.root_event)
    assert not event3.host
    with pytest.raises(ValueError):
        finding = scan.make_event({"path": "/tmp/asdf.txt", "description": "test"}, "FINDING", parent=event3)
    with pytest.raises(ValueError):
        vuln = scan.make_event(
            {"path": "/tmp/asdf.txt", "description": "test", "severity": "HIGH"}, "VULNERABILITY", parent=event3
        )
